from typing import Union


from ..utils.decorators import geokube_logging
import numpy as np

from .axis import Axis, AxisType
from .coordinate import Coordinate, CoordinateType
from .enums import LatitudeConvention, LongitudeConvention


class DomainMixin:
    def __init__():
        pass

    @property
    def latitude(self):
        return self[AxisType.LATITUDE]

    @property
    def longitude(self):
        return self[AxisType.LONGITUDE]

    @property
    def vertical(self):
        return self[AxisType.VERTICAL]

    @property
    def time(self):
        return self[AxisType.TIME]

    @property
    def x(self):
        return self[AxisType.X]

    @property
    def y(self):
        return self[AxisType.Y]

    @property
    def longitude_convention(self) -> LongitudeConvention:
        if np.all(self.longitude.values >= 0.0) and np.all(
            self.longitude.values <= 360.0
        ):
            return LongitudeConvention.POSITIVE_WEST
        if np.all(self.longitude.values >= -180.0) and np.all(
            self.longitude.values <= 180.0
        ):
            return LongitudeConvention.NEGATIVE_WEST
        raise ValueError("Wrong longitude values")

    @property
    def latitude_convention(self) -> LatitudeConvention:
        if AxisType.LATITUDE in self._axis_to_name:
            return self[AxisType.LATITUDE].convention

    @property
    def is_latitude_independent(self):
        return self[AxisType.LATITUDE].type is CoordinateType.INDEPENDENT

    @property
    def is_longitude_independent(self):
        return self[AxisType.LONGITUDE].type is CoordinateType.INDEPENDENT

    @geokube_logging
    def __getitem__(self, key: Union[AxisType, Axis, str]) -> Coordinate:
        if isinstance(key, str):
            return self.coords[key]
        elif isinstance(key, AxisType):
            if key not in self._axis_to_name:
                raise KeyError(
                    f"Axis of type `{key}` does not exist in the domain!"
                )
            return self.coords[self._axis_to_name.get(key)]
        elif isinstance(key, Axis):
            if key.type not in self._axis_to_name:
                raise KeyError(
                    f"Axis of type `{key}` does not exist in the domain!"
                )
            return self.coords[self._axis_to_name.get(key.type)]
        raise TypeError(
            "Indexing coordinates for Domain is supported only for object of"
            " types [str, geokube.Axis, geokub.AxisType]. Provided type:"
            f" {type(key)}"
        )

    @geokube_logging
    def get(self, key: Union[AxisType, Axis, str], default=None):
        if isinstance(key, str):
            return self.coords.get(key, default)
        elif isinstance(key, AxisType):
            if key in self._axis_to_name:
                return self.coords[self._axis_to_name.get(key)]
            return default
        elif isinstance(key, Axis):
            if key.type in self._axis_to_name:
                return self.coords[self._axis_to_name.get(key.type)]
            return default
        raise TypeError(
            "Indexing coordinates for Domain is supported only for object of"
            " types [str, geokube.Axis, geokub.AxisType]. Provided type:"
            f" {type(key)}"
        )



from __future__ import annotations

import warnings
from html import escape
from numbers import Number
from string import Formatter, Template
from typing import (
    Any,
    Hashable,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import dask.array as da
import numpy as np
import xarray as xr
import pandas as pd
from xarray.core.options import OPTIONS

from ..utils import formatting, formatting_html, util_methods
from ..utils.decorators import geokube_logging
from ..utils.hcube_logger import HCubeLogger
from .axis import Axis, AxisType
from .unit import Unit


class Variable(xr.Variable):
    __slots__ = (
        "_dimensions",
        "_units",
    )

    _LOG = HCubeLogger(name="Variable")

    def __init__(
        self,
        data: Union[np.ndarray, da.Array, xr.Variable, Number, Variable],
        dims: Optional[Union[Tuple[Axis], Tuple[AxisType], Tuple[str]]] = None,
        units: Optional[Union[Unit, str]] = None,
        properties: Optional[Mapping[Hashable, str]] = None,
        encoding: Optional[Mapping[Hashable, str]] = None,
    ):
        if isinstance(data, pd.core.indexes.datetimes.DatetimeIndex):
            data = np.array(data)
        if not (
            isinstance(data, np.ndarray)
            or isinstance(data, da.Array)
            or isinstance(data, Variable)
            or isinstance(data, Number)
        ):
            raise TypeError(
                "Expected argument is one of the following types"
                " `number.Number`, `numpy.ndarray`, `dask.array.Array`, or"
                f" `xarray.Variable`, but provided {type(data)}"
            )
        _is_scalar = False
        if isinstance(data, Number):
            data = np.array(data, ndmin=1)
            _is_scalar = True
        if isinstance(data, Variable):
            self._dimensions = data._dimensions
            self._units = data._units
            super().__init__(
                data=data.data,
                dims=data.dim_names,
                attrs=data.properties,
                encoding=data.encoding,
            )
        else:
            self._dimensions = None
            if dims is not None:
                dims = self._as_dimension_tuple(dims)
                dims = np.array(dims, ndmin=1, dtype=Axis)
                if (not _is_scalar) and len(dims) != data.ndim:
                    raise ValueError(
                        f"Provided data have {data.ndim} dimension(s) but"
                        f" {len(dims)} Dimension(s) provided in `dims`"
                        " argument"
                    )

                self._dimensions = dims
            # xarray.Variable must be created with non-None `dims`
            super().__init__(
                data=data,
                dims=self.dim_names,
                attrs=properties,
                encoding=encoding,
                fastpath=True,
            )
            self._units = (
                Unit(units)
                if isinstance(units, str) or units is None
                else units
            )

    def _as_dimension_tuple(self, dims) -> Tuple[Axis, ...]:
        if isinstance(dims, str):
            return (Axis(dims, is_dim=True),)
        elif isinstance(dims, Axis):
            return (dims,)
        elif isinstance(dims, AxisType):
            return (Axis(dims.axis_type_name, axistype=dims, is_dim=True),)
        elif isinstance(dims, Iterable):
            _dims = []
            for d in dims:
                if isinstance(d, str):
                    _dims.append(Axis(name=d, is_dim=True))
                elif isinstance(d, AxisType):
                    _dims.append(
                        Axis(name=d.axis_type_name, axistype=d, is_dim=True)
                    )
                elif isinstance(d, Axis):
                    _dims.append(d)
                else:
                    raise TypeError(
                        "Expected argument of collection item is one of the"
                        " following types `str` or `geokube.Axis`, but"
                        f" provided {type(d)}"
                    )
            return tuple(_dims)
        raise ValueError(
            "Expected argument is one of the following types `str`, `iterable"
            " of str`, `iterable of geokub.Axis`, or `iterable of str`, but"
            f" provided {type(dims)}"
        )

    @property
    def dims(self) -> Tuple[Axis, ...]:
        return self._dimensions

    @property
    def dim_names(self):
        return (
            tuple([d.name for d in self._dimensions])
            if self._dimensions is not None
            else ()
        )

    @property
    def dim_ncvars(self):
        return (
            tuple([d.ncvar for d in self._dimensions])
            if self._dimensions is not None
            else ()
        )

    @property
    def properties(self):
        return self.attrs

    @property
    def units(self) -> Unit:
        return self._units

    def __repr__(self) -> str:
        return self.to_xarray(encoding=False).__repr__()

    def _repr_html_(self):
        return self.to_xarray(encoding=False)._repr_html_()

    def convert_units(self, unit, inplace=True):
        unit = Unit(unit) if isinstance(unit, str) else unit
        if not isinstance(self.data, np.ndarray):
            Variable._LOG.warn(
                "Converting units is supported only for np.ndarray inner data"
                " type. Data will be loaded into the memory!"
            )
            self.data = np.array(
                self.data
            )  # TODO: inplace for cf.Unit doesn't work!
        res = self.units.convert(self.data, unit, inplace)
        if not inplace:
            return Variable(
                data=res,
                dims=self.dims,
                units=unit,
                properties=self.properties,
                encoding=self.encoding,
            )
        self.data = res
        self.units = unit

    @classmethod
    @geokube_logging
    def _get_name(
        cls,
        da: Union[xr.Dataset, xr.DataArray],
        mapping: Optional[Mapping[Hashable, str]],
        id_pattern: str,
    ) -> str:
        if mapping is not None and da.name in mapping:
            return mapping[da.name].get("name", da.name)
        if id_pattern is None:
            return da.attrs.get("standard_name", da.name)
        fmt = Formatter()
        _, field_names, _, _ = zip(*fmt.parse(id_pattern))
        field_names = [f for f in field_names if f]
        # Replace intake-like placeholder to string.Template-like ones
        for k in field_names:
            if k not in da.attrs:
                warnings.warn(
                    f"Requested id_pattern component - `{k}` is not present"
                    " among provided attributes!"
                )
                return da.name
            id_pattern = id_pattern.replace(
                f"{{{k}}}", f"${{{k}}}"
            )  # "{some_field}" -> "${some_field}"
        template = Template(id_pattern)
        return template.substitute(**da.attrs)

    @classmethod
    @geokube_logging
    def from_xarray(
        cls,
        da: xr.DataArray,
        id_pattern: Optional[str] = None,
        copy: Optional[bool] = False,
        mapping: Optional[Mapping[str, Mapping[str, str]]] = None,
    ):
        if not isinstance(da, xr.DataArray):
            raise TypeError(
                "Expected argument of the following type `xarray.DataArray`,"
                f" but provided {type(da)}"
            )
        data = da.data.copy() if copy else da.data
        dims = []
        for d in da.dims:
            if d in da.coords:
                d_name = Variable._get_name(da[d], mapping, id_pattern)
                # If id_pattern is defined, AxisType might be improperly parsed (to GENERIC)
                d_axis = da[d].attrs.get("axis", AxisType.parse(d))
                dims.append(
                    Axis(
                        name=d_name,
                        axistype=d_axis,
                        encoding={"name": da[d].encoding.get("name", d)},
                        is_dim=True,
                    )
                )
            else:
                dims.append(Axis(name=d, is_dim=True))

        dims = tuple(dims)
        attrs = da.attrs.copy()
        encoding = da.encoding.copy()

        units = Unit(
            encoding.pop("units", attrs.pop("units", None)),
            calendar=encoding.pop("calendar", attrs.pop("calendar", None)),
        )

        return Variable(
            data=data,
            dims=dims,
            units=units,
            properties=attrs,
            encoding=encoding,
        )

    @geokube_logging
    def to_xarray(self, encoding=True) -> xr.Variable:
        nc_attrs = self.properties
        nc_encoding = self.encoding
        if encoding:
            dims = self.dim_ncvars
        else:
            dims = self.dim_names
        if self.units is not None and not self.units.is_unknown:
            if self.units.is_time_reference():
                nc_encoding["units"] = self.units.cftime_unit
                nc_encoding["calendar"] = self.units.calendar
            elif np.issubdtype(self.dtype, np.timedelta64) or np.issubdtype(
                self.dtype, np.datetime64
            ):
                # NOTE: issue while using xarray.to_netcdf if units
                # are stored as attributes,
                # example: fapar/10-daily/LENGTH_AFTER
                nc_encoding["units"] = str(self.units)
            else:
                nc_attrs["units"] = str(self.units)

        return xr.Variable(
            data=self._data,
            dims=dims,
            attrs=nc_attrs,
            encoding=nc_encoding,
            fastpath=True,
        )



from __future__ import annotations

import re
from enum import Enum
from typing import Any, Hashable, List, Mapping, Optional, Union

import xarray as xr

from .unit import Unit
from ..utils.hcube_logger import HCubeLogger

# from https://unidata.github.io/MetPy/latest/_modules/metpy/xarray.html
coordinate_criteria_regular_expression = {
    "y": r"(y|rlat|grid_lat.*|j|projection_y_coordinate)",
    "x": r"(x|rlon|grid_lon.*|i|projection_x_coordinate)",
    "vertical": r"(soil|lv_|bottom_top|sigma|h(ei)?ght|altitude|dept(h)?|isobaric|pres|isotherm|model_level_number)[a-z_]*[0-9]*",
    "timedelta": r"time_delta",
    "time": r"(time[0-9]*|T)",
    "latitude": r"(x?lat[a-z0-9]*|nav_lat)",
    "longitude": r"(x?lon[a-z0-9]*|nav_lon)",
}


class AxisType(Enum):
    TIME = ("time", Unit("hours since 1970-01-01", calendar="gregorian"))
    TIMEDELTA = ("timedelta", Unit("hour"))
    LATITUDE = ("latitude", Unit("degrees_north"))
    LONGITUDE = ("longitude", Unit("degrees_east"))
    VERTICAL = ("vertical", Unit("m"))
    X = ("x", Unit("m"))
    Y = ("y", Unit("m"))
    Z = ("z", Unit("m"))
    RADIAL_AXIMUTH = ("aximuth", Unit("m"))
    RADIAL_ELEVATION = ("elevation", Unit("m"))
    RADIAL_DISTANCE = ("distance", Unit("m"))
    GENERIC = ("generic", Unit("Unknown"))

    @property
    def default_unit(self) -> Unit:
        return self.value[1]

    @property
    def axis_type_name(self) -> str:
        return self.value[0]

    @classmethod
    def values(cls) -> List[str]:
        return [a.value[1] for a in cls]

    @classmethod
    def parse(cls, name) -> "AxisType":
        if name is None:
            return cls.GENERIC
        if isinstance(name, AxisType):
            return name
        if isinstance(name, Axis):
            return name.type
        try:
            res = cls[name.upper() if isinstance(name, str) else name]
            if res is AxisType.Z:
                return AxisType.VERTICAL
            return res
        except KeyError:
            for ax, regexp in coordinate_criteria_regular_expression.items():
                if re.match(regexp, name.lower(), re.IGNORECASE):
                    return cls[ax.upper()]
        return cls.GENERIC

    @classmethod
    def _missing_(cls, key) -> "AxisType":
        return cls.GENERIC


class Axis:
    _LOG = HCubeLogger(name="Axis")

    def __init__(
        self,
        name: Union[str, Axis],
        axistype: Optional[Union[AxisType, str]] = None,
        encoding: Optional[Mapping[Hashable, str]] = None,
        is_dim: Optional[bool] = False,
    ):
        if isinstance(name, Axis):
            self._name = name._name
            self._type = name._type
            self._encoding = name._encoding
            self._is_dim = name._is_dim
        else:
            self._is_dim = is_dim
            self._name = name
            self._encoding = encoding
            if axistype is None:
                self._type = AxisType.parse(name)
            else:
                if isinstance(axistype, str):
                    self._type = AxisType.parse(axistype)
                elif isinstance(axistype, AxisType):
                    self._type = axistype
                else:
                    raise TypeError(
                        "Expected argument is one of the following types"
                        " `str`, `geokube.AxisType`, but provided"
                        f" {type(axistype)}"
                    )

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> AxisType:
        return self._type

    @property
    def default_unit(self) -> Unit:
        return self._type.default_unit

    @property
    def ncvar(self):
        return (
            self._encoding.get("name", self.name)
            if self._encoding
            else self.name
        )

    @property
    def encoding(self):
        return self._encoding

    @property
    def is_dim(self):
        return self._is_dim

    def __hash__(self):
        enc_keys = (
            tuple(self._encoding.keys())
            if self._encoding is not None
            else tuple()
        )
        return hash((self._name, self._type, enc_keys, self._is_dim))

    def __eq__(self, other):
        if issubclass(type(other), Axis):
            return (
                (self.name == other.name)
                and (self.type == other.type)
                and (self._is_dim == other.is_dim)
                and (self._encoding == other._encoding)
            )
        return False

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self) -> str:
        return (
            f"<Axis(name={self.name}, type:{self.type},"
            f" encoding={self._encoding}>"
        )

    def __str__(self) -> str:
        return f"{self.name}: {self.type}"

    @classmethod
    def get_name_for_object(cls, obj: Union[str, Axis, AxisType]) -> str:
        if isinstance(obj, Axis):
            return obj.name
        elif isinstance(obj, str):
            return obj
        elif isinstance(obj, AxisType):
            return obj.axis_type_name
        else:
            raise TypeError(
                "`dims` can be a tuple or a list of [geokube.Axis,"
                f" geokube.AxisType, str], but provided type is `{type(obj)}`"
            )



"""geokube framework"""
import setuptools
from distutils.util import convert_path

main_ns = {}
ver_path = convert_path("geokube/version.py")
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="geokube",
    version=main_ns["__version__"],
    author="geokube Contributors",
    author_email="geokube@googlegroups.com",
    description="a Python package based on xarray for GeoScience Data Analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/CMCC-Foundation/geokube",
    packages=setuptools.find_packages(),
    install_requires=[line.strip() for line in open("requirements.txt").readlines()],
    #install_requires=[
    #    "shapely",
    #    "metpy",
    #    "plotly",
    #    "pyarrow",
    #     "rioxarray",
    #],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering",
    ],
    python_requires=">=3.10",
    license="Apache License, Version 2.0",
    package_data={"geokube": ["static/css/*.css", "static/html/*.html"]},
)



class ClassProperty(object):
    def __init__(self, fget):
        self.fget = fget

    def __get__(self, obj, klass=None):
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()


def classproperty(func):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassProperty(func)
